using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using AssetBundles;
using SimpleJSON;

namespace dubPlugin {
	//<summary>
	// Original plugins by Extraltodeus (PumpTrigger) and MeshedVR (TriggerControl)
	// Pimped by dub
	//</summary>
	public class PimpedPumpTrigger : MVRScript, TriggerHandler {
		
		private JSONStorableFloat _pumpValueJSON;
		private Trigger _trigger;

		public override void Validate() {
			base.Validate();
			_trigger?.Validate();
		}
		
		public void RemoveTrigger(Trigger triggerParam) {
			// do nothing as we only manage single trigger here
		}

		public void DuplicateTrigger(Trigger triggerParam) {
			// do nothing as we only manage single trigger here
		}

		private void SyncTriggerActive(bool b) {
			if (_trigger != null && _enableTriggerJSON.val) {
				_trigger.active = b;
			}
		}

		public void SyncVariableControl(float f) {
			if (_trigger != null && _enableTriggerJSON.val) {
				_trigger.transitionInterpValue = f;
			}
		}
		
		// this function allows trigger atom references to be updated when an atom is renamed
		private void OnAtomRename(string oldId, string newId)
		{
			_trigger?.SyncAtomNames();
		}

		private void OpenTriggerActionsPanel() {
			_trigger.OpenTriggerActionsPanel();
		}

		private JSONStorableAction _openTriggerActionsPanelAction;

		private void SyncAtomChoices() {
			var atomChoices = new List<string> {"None"};
			atomChoices.AddRange(SuperController.singleton.GetAtomUIDs());
			_atomJSON.choices = atomChoices;
		}

		// receiver Atom
		private Atom _receivingAtom;

		private void SyncAtom(string atomUID) {
			var receiverChoices = new List<string> {"None"};
			if (atomUID != null) {
				_receivingAtom = SuperController.singleton.GetAtomByUid(atomUID);
				if (_receivingAtom != null)
				{
					receiverChoices.AddRange(_receivingAtom.GetStorableIDs());
				}
			} else {
				_receivingAtom = null;
			}
			_receiverJSON.choices = receiverChoices;
			_receiverJSON.val = "None";
		}

		private JSONStorableStringChooser _atomJSON;

		private string _missingReceiverStoreId = "";

		private void CheckMissingReceiver()
		{
			if (_missingReceiverStoreId == "" ||
			    _receivingAtom == null) return;
			var missingReceiver = _receivingAtom.GetStorableByID(_missingReceiverStoreId);
			if (missingReceiver == null) return;
			var saveTargetName = _receiverTargetName;
			SyncReceiver(_missingReceiverStoreId);
			_missingReceiverStoreId = "";
			insideRestore = true;
			_receiverTargetJSON.val = saveTargetName;
			insideRestore = false;
		}
		
		private JSONStorable _receiver;

		private void SyncReceiver(string receiverID) {
			var receiverTargetChoices = new List<string> {"None"};
			if (_receivingAtom != null && receiverID != null) {
				_receiver = _receivingAtom.GetStorableByID(receiverID);
				if (_receiver != null)
				{
					receiverTargetChoices.AddRange(_receiver.GetFloatParamNames());
				} else if (receiverID != "None") {
					_missingReceiverStoreId = receiverID;
				}
			} else {
				_receiver = null;
			}
			_receiverTargetJSON.choices = receiverTargetChoices;
			_receiverTargetJSON.val = "None";
		}

		private JSONStorableStringChooser _receiverJSON;

		private string _receiverTargetName;
		private JSONStorableFloat _receiverTarget;

		private void SyncReceiverTarget(string receiverTargetNameParam) {
			_receiverTargetName = receiverTargetNameParam;
			_receiverTarget = null;
			if (_receiver != null && receiverTargetNameParam != null) {
				_receiverTarget = _receiver.GetFloatJSONParam(receiverTargetNameParam);
				if (_receiverTarget == null) return;
				_lowerValueJSON.min = _receiverTarget.min;
				_lowerValueJSON.max = _receiverTarget.max;
				_upperValueJSON.min = _receiverTarget.min;
				_upperValueJSON.max = _receiverTarget.max;
				_currentValueJSON.min = _receiverTarget.min;
				_currentValueJSON.max = _receiverTarget.max;
				_targetValueJSON.min = _receiverTarget.min;
				_targetValueJSON.max = _receiverTarget.max;
				if (insideRestore) return;
				// only sync up val if not in restore
				_lowerValueJSON.val = _receiverTarget.val;
				_upperValueJSON.val = _receiverTarget.val;
				_currentValueJSON.val = _receiverTarget.val;
				_targetValueJSON.val = _receiverTarget.val;
			}
		}
		private JSONStorableStringChooser _receiverTargetJSON;

		private JSONStorableBool _enableCooldownJSON;
		private JSONStorableBool _enablePumpCooldownJSON;
		private JSONStorableBool _cooldownActiveJSON;
		private JSONStorableBool _enableTriggerJSON;
		private JSONStorableBool _alwaysSendTransitionJSON;
		private JSONStorableBool _triggerHitJSON;
		private JSONStorableBool _pumpOnce;

		private JSONStorableFloat _triggerThresholdJSON;
		private JSONStorableFloat _cooldownDurationJSON;
		private JSONStorableFloat _downSpeedJSON;
		private JSONStorableFloat _upSpeedJSON;
		private JSONStorableFloat _durationJSON;
		private JSONStorableFloat _lowerValueJSON;
		private JSONStorableFloat _upperValueJSON;
		private JSONStorableFloat _targetValueJSON;
		private JSONStorableFloat _currentValueJSON;
		private JSONStorableFloat _smoothLoopFactor;
		
		private UIDynamicButton _cooldownActiveButton;
		private UIDynamicButton _triggerHitButton;
		private UIDynamicButton _speedController;

		private static readonly Color ActiveTextColor = Color.white;
		private static readonly Color InactiveTextColor = Color.clear;
		private static readonly Color ActiveBgColor = new Color(0.16f, 1f, 0f);
		private static readonly Color InactiveBgColor = new Color(0.23f, 0.23f, 0.23f);
		
		private readonly Dictionary<int, float> _smoothLoops = new Dictionary<int, float>();

		#region Init / UI
		public override void Init() {
			try
			{
				_trigger = new Trigger { handler = this };
				StartCoroutine(LoadUIAssets());

				// handle atom renaming - must be passed to trigger
				if (SuperController.singleton != null) {
					SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
				}
				_openTriggerActionsPanelAction = new JSONStorableAction("OpenTriggerActionsPanel", OpenTriggerActionsPanel);
				var db = CreateButton("Edit Trigger Actions");
				_openTriggerActionsPanelAction.dynamicButton = db;

				_enableTriggerJSON = new JSONStorableBool("Enable Trigger", false, SyncTriggerActive);
				RegisterBool(_enableTriggerJSON);
				CreateToggle(_enableTriggerJSON);
				
				_triggerThresholdJSON = new JSONStorableFloat("Trigger Threshold", 1f, 0f, 1f);
				RegisterFloat(_triggerThresholdJSON);
				CreateSlider(_triggerThresholdJSON);
				
				_alwaysSendTransitionJSON = new JSONStorableBool("Always Set Trigger Transition Value", true);
				RegisterBool(_alwaysSendTransitionJSON);
				CreateToggle(_alwaysSendTransitionJSON);
				
				_enableCooldownJSON = new JSONStorableBool("Enable Trigger Cooldown", true);
				RegisterBool(_enableCooldownJSON);
				CreateToggle(_enableCooldownJSON);

				_cooldownActiveJSON = new JSONStorableBool("Cooldown Active", false, (bool on) =>
				{
					_cooldownTime = _cooldownDurationJSON.val;
					_cooldownActiveButton.buttonColor = Color.green;
				});
				RegisterBool(_cooldownActiveJSON);
				
				_enablePumpCooldownJSON = new JSONStorableBool("Prevent Pumping During Cooldown", false);
				RegisterBool(_enablePumpCooldownJSON);
				CreateToggle(_enablePumpCooldownJSON);
				
				_cooldownDurationJSON = new JSONStorableFloat("Cooldown Duration", 1f, 0f, 60f, false);
				RegisterFloat(_cooldownDurationJSON);
				CreateSlider(_cooldownDurationJSON);
				
				_triggerHitJSON = new JSONStorableBool("Trigger Hit", false, on=>
				{
					_trigger.active = (!_enableCooldownJSON.val || !_cooldownActiveJSON.val) && on;
					if (_enableCooldownJSON.val && _cooldownActiveJSON.val == false) _cooldownActiveJSON.val = true;
				})
				{
					isStorable = false,
					isRestorable = false
				};
				RegisterBool(_triggerHitJSON);
				
				CreateSpacer().height = 20f;
				
				_triggerHitButton = CreateButton("Trigger");
				_triggerHitButton.button.interactable = false;
				_cooldownActiveButton = CreateButton("Cooldown");
				_cooldownActiveButton.button.interactable = false;

				CreateSpacer().height = 10f;
				
				_speedController = CreateButton("Pump");
				_speedController.button.onClick.AddListener(delegate
				{
					_toElapse = _durationJSON.val;
				});

				_pumpOnce = new JSONStorableBool("Pump once", false, (bool on)=>
				{
					_pumpOnce.val = false;
					_toElapse = _durationJSON.val;
				});
				RegisterBool(_pumpOnce);

				_atomJSON = new JSONStorableStringChooser("atom", SuperController.singleton.GetAtomUIDs(), null, "Atom", SyncAtom);
				RegisterStringChooser(_atomJSON);
				SyncAtomChoices();
				#if(VAM_GT_1_20)
				var popupAtom = CreateFilterablePopup(_atomJSON, true);
				#else
				var popupAtom = CreateScrollablePopup(_atomJSON, true);
				#endif
				popupAtom.popupPanelHeight = 1100f;
				// want to always resync the atom choices on opening popup since atoms can be added/removed
				popupAtom.popup.onOpenPopupHandlers += SyncAtomChoices;

				// make receiver selector
				_receiverJSON = new JSONStorableStringChooser("receiver", null, null, "Receiver", SyncReceiver);
				RegisterStringChooser(_receiverJSON);
				#if(VAM_GT_1_20)
				var popupReceiver = CreateFilterablePopup(_receiverJSON,true);
				#else
				var popupReceiver = CreateScrollablePopup(_receiverJSON,true);
				#endif
				popupReceiver.popupPanelHeight = 960f;

				// make receiver target selector
				_receiverTargetJSON = new JSONStorableStringChooser("receiverTarget", null, null, "Target", SyncReceiverTarget);
				RegisterStringChooser(_receiverTargetJSON);
				#if(VAM_GT_1_20)
				var popupReceiverTarget = CreateFilterablePopup(_receiverTargetJSON,true);
				#else
				var popupReceiverTarget = CreateScrollablePopup(_receiverTargetJSON,true);
				#endif
				popupReceiverTarget.popupPanelHeight = 820f;

				// set atom to current atom to initialize
				_atomJSON.val = containingAtom.uid;

				_smoothLoopFactor = new JSONStorableFloat("Smooth Factor", 50f, 1f, 100f);
				RegisterFloat(_smoothLoopFactor);
				CreateSlider(_smoothLoopFactor, true);

				// morphSlider.slider.onValueChanged.AddListener(delegate (float mixValue)


				_downSpeedJSON = new JSONStorableFloat("Down Speed", 0.5f, 0f, 10f);
				RegisterFloat(_downSpeedJSON);
				CreateSlider(_downSpeedJSON, true);

				_upSpeedJSON = new JSONStorableFloat("Up Speed", 1f, 0f, 10f);
				RegisterFloat(_upSpeedJSON);
				CreateSlider(_upSpeedJSON, true);

				_durationJSON = new JSONStorableFloat("Duration", 20f, 1f, 100f);
				RegisterFloat(_durationJSON);
				CreateSlider(_durationJSON, true);

				// lower val
				_lowerValueJSON = new JSONStorableFloat("Min. Value", 0f, 0f, 1f, false);
				RegisterFloat(_lowerValueJSON);
				CreateSlider(_lowerValueJSON, true);

				// upper val
				_upperValueJSON = new JSONStorableFloat("Max. Value", 1f, 0.1f, 1f, false);
				RegisterFloat(_upperValueJSON);
				CreateSlider(_upperValueJSON, true);

				CreateSpacer().height = 10f;

				_pumpValueJSON = new JSONStorableFloat("Pump Value", 0f, 0f, 1f, false, false)
				{
					isStorable = false,
					isRestorable = false
				};
				// don't register - this is for viewing only and is generated
				var sliderPumpValue = CreateSlider(_pumpValueJSON);
				sliderPumpValue.defaultButtonEnabled = false;
				sliderPumpValue.quickButtonsEnabled = false;

				// target value
				_targetValueJSON = new JSONStorableFloat("targetValue", 0f, 0f, 1f, false, false)
				{
					isStorable = false,
					isRestorable = false
				};
				
				// current value
				_currentValueJSON = new JSONStorableFloat("Current Receiver Value", 0f, 0f, 1f, false, false)
				{
					isStorable = false,
					isRestorable = false
				};
				// don't register - this is for viewing only
				var sliderCurrentValue = CreateSlider(_currentValueJSON);
				sliderCurrentValue.defaultButtonEnabled = false;
				sliderCurrentValue.quickButtonsEnabled = false;

			}

			catch (Exception e) {
				SuperController.LogError("Exception caught: " + e);
			}
		}

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			var json = base.GetJSON(includePhysical, includeAppearance, forceStore);
			json["trigger"] = _trigger.GetJSON();
			return json;
		}

		public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true,
			JSONArray presetAtoms = null, bool setMissingToDefault = true)
		{
			base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
			if (jc["trigger"] != null) _trigger.RestoreFromJSON(jc["trigger"].AsObject);
		}

		public override void InitUI()
		{
			base.InitUI();
			if (UITransform != null) {
				_trigger.triggerActionsParent = UITransform;
			}
		}
		#endregion
		
		#region Lifecycle
		private int _currentLoop;
		
		private float _toElapse;
		private float _pump;
		private float _speed;
		private float _cooldownTime;
		
		private bool _isHit;
		private bool _cooldown;

		protected void FixedUpdate()
		{
			_cooldown = _enableCooldownJSON.val && _cooldownActiveJSON.val;
			if (_cooldown)
			{
				if (_cooldownTime <= 0)
				{
					_cooldownTime = 0;
					_cooldownActiveJSON.val = false;
				}
				else
				{
					_cooldownTime -= Time.deltaTime;
				}
			}

			_cooldownActiveButton.textColor = _cooldown ? ActiveTextColor : InactiveTextColor;
			_cooldownActiveButton.buttonColor = _cooldown ? ActiveBgColor : InactiveBgColor;

			if(_toElapse > 0)
			{
				if (_cooldown == false || _enablePumpCooldownJSON.val == false)
				{
					_speed = _upSpeedJSON.val;
				}
				_toElapse--;
			}
			else
			{
				_speed = -_downSpeedJSON.val;
			}

			if (_toElapse<=0)
			{
				_toElapse = 0;
			}

			_pump += _speed/100f;
			_isHit = _cooldown == false && _pump >= _triggerThresholdJSON.val;

			if (_pump>1f)
			{
				_pump = 1f;
			}

			if (_pump<0)
			{
				_pump = 0;
			}

			_pumpValueJSON.val = _pump;
			_triggerHitJSON.val = _enableTriggerJSON.val && !_cooldown && _isHit;
			SyncTriggerActive(_triggerHitJSON.val);
			if (_alwaysSendTransitionJSON.val || _triggerHitJSON.val)
			{
				SyncVariableControl(_pump);
			}

			_triggerHitButton.textColor = _triggerHitJSON.val ? ActiveTextColor : InactiveTextColor;
			_triggerHitButton.buttonColor = _triggerHitJSON.val ? ActiveBgColor : InactiveBgColor;

				if (_receiverTarget != null) {

				int smoothFactor = (int)Math.Ceiling(_smoothLoopFactor.val);


				var result = _pump * (_upperValueJSON.val - _lowerValueJSON.val) + _lowerValueJSON.val ;

				if (result > _upperValueJSON.val)
					result = _upperValueJSON.val;

				if (result < _lowerValueJSON.val)
					result = _lowerValueJSON.val;
					
				_smoothLoops[_currentLoop] = result;

				int divider = 0;
				float total = 0;
				for (var i = 0; i < smoothFactor; i++)
				{
					if (_smoothLoops.ContainsKey(i))
					{
						divider++;
						total = total+_smoothLoops[i];
					}
					else
					{
						_smoothLoops[i] = 0;
						divider++;
					}
				}

				result = total/divider;
				_currentValueJSON.val = result;

				if (_currentLoop < smoothFactor)
				{
					_currentLoop++;
				}
				else
				{
					_currentLoop = 0;
				}
			}

			// check for receivers that might have been missing on load due to asynchronous load of some assets like skin, clothing, hair
			CheckMissingReceiver();
			if (_receiverTarget != null) {
				_receiverTarget.val  = _currentValueJSON.val;
			}
		}
		#endregion
		
		#region Callbacks
		
		protected void OnDestroy()
		{
			if (SuperController.singleton == null) return;
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
		}
		
		#endregion
		
		#region Trigger Action UI
		
		public RectTransform CreateTriggerActionsUI() {
			RectTransform rt = null;
			if (_triggerActionsPrefab != null) {
				rt = Instantiate(_triggerActionsPrefab);
			} else {
				Debug.LogError("Attempted to make TriggerActionsUI when prefab was not set");
			}
			return (rt);
		}

		public RectTransform CreateTriggerActionMiniUI() {
			RectTransform rt = null;
			if (_triggerActionMiniPrefab != null) {
				rt = Instantiate(_triggerActionMiniPrefab);
			} else {
				Debug.LogError("Attempted to make TriggerActionMiniUI when prefab was not set");
			}
			return (rt);
		}

		public RectTransform CreateTriggerActionDiscreteUI() {
			RectTransform rt = null;
			if (_triggerActionDiscretePrefab != null) {
				rt = Instantiate(_triggerActionDiscretePrefab);
			} else {
				Debug.LogError("Attempted to make TriggerActionDiscreteUI when prefab was not set");
			}
			return (rt);
		}

		public RectTransform CreateTriggerActionTransitionUI() {
			RectTransform rt = null;
			if (_triggerActionTransitionPrefab != null) {
				rt = Instantiate(_triggerActionTransitionPrefab);
			} else {
				Debug.LogError("Attempted to make TriggerActionTransitionUI when prefab was not set");
			}
			return (rt);
		}

		public void RemoveTriggerActionUI(RectTransform rt) {
			if (rt != null) {
				Destroy(rt.gameObject);
			}
		}
		
		#endregion
		
		#region Asset Loading

		private RectTransform _triggerActionsPrefab;
		private RectTransform _triggerActionMiniPrefab;
		private RectTransform _triggerActionDiscretePrefab;
		private RectTransform _triggerActionTransitionPrefab;
		
		private IEnumerator LoadUIAssets() {
			var request = AssetBundleManager.LoadAssetAsync("z_ui2", "TriggerActionsPanel", typeof(GameObject));
			if (request == null) {
				SuperController.LogError("Request for TriggerActionsPanel in z_ui2 assetbundle failed");
				yield break;
			}
			yield return request;
			var go = request.GetAsset<GameObject>();
			if (go == null) {
				SuperController.LogError("Failed to load TriggerActionsPanel asset");
			}
			_triggerActionsPrefab = go.GetComponent<RectTransform>();
			if (_triggerActionsPrefab == null) {
				SuperController.LogError("Failed to load TriggerActionsPanel asset");
			}

			request = AssetBundleManager.LoadAssetAsync("z_ui2", "TriggerActionMiniPanel", typeof(GameObject));
			if (request == null) {
				SuperController.LogError("Request for TriggerActionMiniPanel in z_ui2 assetbundle failed");
				yield break;
			}
			yield return request;
			go = request.GetAsset<GameObject>();
			if (go == null) {
				SuperController.LogError("Failed to load TriggerActionMiniPanel asset");
			}
			_triggerActionMiniPrefab = go.GetComponent<RectTransform>();
			if (_triggerActionMiniPrefab == null) {
				SuperController.LogError("Failed to load TriggerActionMiniPanel asset");
			}

			request = AssetBundleManager.LoadAssetAsync("z_ui2", "TriggerActionDiscretePanel", typeof(GameObject));
			if (request == null) {
				SuperController.LogError("Request for TriggerActionDiscretePanel in z_ui2 assetbundle failed");
				yield break;
			}
			yield return request;
			go = request.GetAsset<GameObject>();
			if (go == null) {
				SuperController.LogError("Failed to load TriggerActionDiscretePanel asset");
			}
			_triggerActionDiscretePrefab = go.GetComponent<RectTransform>();
			if (_triggerActionDiscretePrefab == null) {
				SuperController.LogError("Failed to load TriggerActionDiscretePanel asset");
			}

			request = AssetBundleManager.LoadAssetAsync("z_ui2", "TriggerActionTransitionPanel", typeof(GameObject));
			if (request == null) {
				SuperController.LogError("Request for TriggerActionTransitionPanel in z_ui2 assetbundle failed");
				yield break;
			}
			yield return request;
			go = request.GetAsset<GameObject>();
			if (go == null) {
				SuperController.LogError("Failed to load TriggerActionTransitionPanel asset");
			}
			_triggerActionTransitionPrefab = go.GetComponent<RectTransform>();
			if (_triggerActionTransitionPrefab == null) {
				SuperController.LogError("Failed to load TriggerActionTransitionPanel asset");
			}

		}
		#endregion


	}
}
